#include <e.h>
#include <Ecore_X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include "e_mod_main.h"

#define SYS_TRAY_REQUEST_DOCK 0
#define SYS_TRAY_BEGIN_MESSAGE 1
#define SYS_TRAY_CANCEL_MESSAGE 2
#define _NET_SYSTEM_TRAY_ORIENTATION_HORZ 0
#define _NET_SYSTEM_TRAY_ORIENTATION_VERT 1

#define XEMBED_EMBEDDED_NOTIFY 0

typedef struct _Instance Instance;
struct _Instance 
{
   E_Gadcon_Client *gcc;
   Evas_Object *o_base, *o_box;
   E_Container *con;
   Evas *evas;
   Ecore_X_Window win;
   E_Menu *menu;
   Ecore_Event_Handler *msg_hdl, *dst_hdl;
   struct 
     {
       int x;
       int y;
     } padding;
   Ecore_X_Pixmap bg_pixmap;
   Evas_List *icons;
};

typedef struct _Tray_Icon Tray_Icon;
struct _Tray_Icon
{
  Ecore_X_Window parent;
  Ecore_X_Window win;
};

/* local function protos */
static E_Gadcon_Client *_gc_init(E_Gadcon *gc, const char *name, const char *id, const char *style);
static void _gc_shutdown(E_Gadcon_Client *gcc);
static void _gc_orient(E_Gadcon_Client *gcc);
static char *_gc_label(void);
static Evas_Object *_gc_icon(Evas *evas);
static const char *_gc_id_new(void);
static void _cb_mouse_down(void *data, Evas *evas, Evas_Object *obj, void *event_info);
static int _cb_msg(void *data, int type, void *event);
static int _cb_destroy(void *data, int type, void *event);
static void _tray_add_icons(void *data);
static void _tray_add_icon(void *data, Ecore_X_Window win);
static void _tray_set_active(void *data, int active);

/* local vars */
static Evas_List *instances = NULL;
static E_Module *module = NULL;

static const E_Gadcon_Client_Class _gc_class = 
{
   GADCON_CLIENT_CLASS_VERSION, "systray", 
     {_gc_init, _gc_shutdown, _gc_orient, _gc_label, _gc_icon, _gc_id_new, NULL}, 
   E_GADCON_CLIENT_STYLE_PLAIN
};

static E_Gadcon_Client *
_gc_init(E_Gadcon *gc, const char *name, const char *id, const char *style) 
{
   Instance *inst = NULL;
   E_Container *con = NULL;
   char buf[4096];
   int x, y, w, h;

   inst = E_NEW(Instance, 1);
   if (!inst) return NULL;
   inst->evas = gc->evas;

   con = e_container_current_get(e_manager_current_get());
   inst->con = con;

   snprintf(buf, sizeof(buf), "%s/systray.edj", 
	    e_module_dir_get(module));
   inst->o_base = edje_object_add(gc->evas);
   if (!e_theme_edje_object_set(inst->o_base, "base/theme/modules/systray", 
				"modules/systray/main"))
     edje_object_file_set(inst->o_base, buf, "modules/systray/main");
   evas_object_show(inst->o_base);

//   inst->o_box = e_box_add(gc->evas);
//   edje_object_part_swallow(inst->o_base, "e.swallow.tray", inst->o_box);
//   evas_object_show(inst->o_box);
//   e_box_align_set(inst->o_box, 0.5, 0.5);

   inst->gcc = e_gadcon_client_new(gc, name, id, style, inst->o_base);
   inst->gcc->data = inst;

   evas_object_event_callback_add(inst->o_base, EVAS_CALLBACK_MOUSE_DOWN, 
				  _cb_mouse_down, inst);

   inst->padding.x = 0;
   inst->padding.y = 0;
   if (gc->shelf->popup)
       inst->win = gc->shelf->win;
   else
       inst->win = con->bg_win;
   /* tell system that we will be the tray now */
   _tray_set_active(inst, 1);

   /* create orig owner win */
   //ecore_x_window_geometry_get(gc->shelf->win, &x, &y, &w, &h);

   /*
   ecore_x_window_container_manage(inst->win);
   inst->win = ecore_x_window_new(gc->shelf->win, x, y, w, h);
   ecore_x_window_background_color_set(inst->win, 0xcccc, 0xcccc, (unsigned short)0x65536);
   ecore_x_window_pixmap_set(inst->win, None);
   XClearWindow(ecore_x_display_get(), inst->win);
   ecore_x_window_show(inst->win);
   */
   /* setup handlers */
   inst->msg_hdl = ecore_event_handler_add(ECORE_X_EVENT_CLIENT_MESSAGE, 
					   _cb_msg, inst);
   inst->dst_hdl = ecore_event_handler_add(ECORE_X_EVENT_WINDOW_DESTROY, 
					   _cb_destroy, inst);
   _tray_add_icons(inst);

   instances = evas_list_append(instances, inst);
   return inst->gcc;
}

static void 
_gc_shutdown(E_Gadcon_Client *gcc) 
{
   Instance *inst = NULL;
   Evas_List *l;

   inst = gcc->data;
   if (!inst) return;
   instances = evas_list_remove(instances, inst);
   _tray_set_active(inst, 0);
   for (l = inst->icons; l; l = l->next)
     {
       Tray_Icon *ico = l->data;
       inst->icons = evas_list_remove_list(inst->icons, l);
       ecore_x_window_del(ico->win);
       ecore_x_window_del(ico->parent);
       free(ico);
     }
   if (inst->bg_pixmap) ecore_x_pixmap_del(inst->bg_pixmap);
   if (inst->msg_hdl) ecore_event_handler_del(inst->msg_hdl);
   if (inst->dst_hdl) ecore_event_handler_del(inst->dst_hdl);
   //if (inst->win) ecore_x_window_del(inst->win);
   if (inst->o_box) evas_object_del(inst->o_box);
   if (inst->o_base) 
     {
	evas_object_event_callback_del(inst->o_base, EVAS_CALLBACK_MOUSE_DOWN, 
				       _cb_mouse_down);
	evas_object_del(inst->o_base);
     }
   E_FREE(inst);
}

static void 
_gc_orient(E_Gadcon_Client *gcc) 
{
   e_gadcon_client_min_size_set(gcc, 16, 16);
}

static char *
_gc_label(void) 
{
   return "Systray";
}

static Evas_Object *
_gc_icon(Evas *evas) 
{
   Evas_Object *o = NULL;
   char buf[4096];

   snprintf(buf, sizeof(buf), "%s/e-module-systray.edj", 
	    e_module_dir_get(module));
   o = edje_object_add(evas);
   edje_object_file_set(o, buf, "icon");
   return o;
}

static const char *
_gc_id_new(void) 
{
   char buf[4096];
   
   snprintf(buf, sizeof(buf), "%s.%d", _gc_class.name, 
	    evas_list_count(instances));
   return strdup(buf);
}

/* E Module API Functions */
EAPI E_Module_Api e_modapi = 
{
   E_MODULE_API_VERSION, "Systray"
};

EAPI void *
e_modapi_init(E_Module *m) 
{
   module = m;
   e_gadcon_provider_register(&_gc_class);
   return m;
}

EAPI int 
e_modapi_shutdown(E_Module *m) 
{
   e_gadcon_provider_unregister(&_gc_class);
   module = NULL;
   return 1;
}

EAPI int 
e_modapi_save(E_Module *m) 
{
   return 1;
}

/* local functions */
static void 
_cb_mouse_down(void *data, Evas *evas, Evas_Object *obj, void *event_info) 
{
   Instance *inst = NULL;
   Evas_Event_Mouse_Down *ev;

   if (!(inst = data)) return;
   ev = event_info;
   if ((ev->button == 3) && (!inst->menu)) 
     {
	E_Zone *zone;
	E_Menu *mn;
	int x, y;

	zone = e_util_zone_current_get(e_manager_current_get());
	mn = e_menu_new();
	e_gadcon_client_util_menu_items_append(inst->gcc, mn, 0);
	e_gadcon_canvas_zone_geometry_get(inst->gcc->gadcon, &x, &y, NULL, NULL);
	e_menu_activate_mouse(mn, zone, x + ev->output.x, y + ev->output.y, 
			      1, 1, E_MENU_POP_DIRECTION_AUTO, ev->timestamp);
	evas_event_feed_mouse_up(inst->gcc->gadcon->evas, ev->button, 
				 EVAS_BUTTON_NONE, ev->timestamp, NULL);
     }
}

static void 
_tray_add_icons(void *data) 
{
   Ecore_X_Window *wins;
   Ecore_X_Atom axmbed, awin, akde, akdock;
   Instance *inst = NULL;
   int num, i;

   if (!(inst = data)) return;
   wins = ecore_x_window_children_get(inst->con->manager->root, &num);
   if (!wins) return;
   awin = ecore_x_atom_get("WINDOW");
   axmbed = ecore_x_atom_get("_XEMBED_INFO");
   akde = ecore_x_atom_get("_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR");
   akdock = ecore_x_atom_get("KWM_DOCKWINDOW");
   for (i = 0; i < num; i++) 
     {
	Ecore_X_Window_Attributes att;
	unsigned char *data = NULL;
	int count;

	if (!ecore_x_window_attributes_get(wins[i], &att)) continue;
	if (!ecore_x_window_prop_property_get(wins[i], axmbed, axmbed, 32, &data, &count))
	  data = NULL;
	if (!data) 
	  {
	     if (!ecore_x_window_prop_property_get(wins[i], akde, akde, 32, &data, &count))
	       data = NULL;
	  }
	if (!data) 
	  {
	     if (!ecore_x_window_prop_property_get(wins[i], akdock, akdock, 32, &data, &count))
	       data = NULL;
	  }
	if (!data) continue;
	printf("Systray Found: %x\n", wins[i]);
	_tray_add_icon(inst, wins[i]);
	free(data);
	data = NULL;
     }
   free(wins);
}

static void 
_tray_add_icon(void *data, Ecore_X_Window win) 
{
   Instance *inst = NULL;
   Tray_Icon *ico;
   int x, y, w, h, b, d, num_ico;

   if (!(inst = data)) return;

   ico = malloc(sizeof(Tray_Icon));
   if (ico == NULL) return;
   ico->parent = ecore_x_window_new(inst->win, 0, 0, 48, 48);
   ico->win = win;
   inst->icons = evas_list_append(inst->icons, ico);
   num_ico = evas_list_count(inst->icons);

   evas_object_geometry_get(inst->o_base, &(inst->padding.x), &(inst->padding.y),
                            NULL, NULL);
   ecore_x_window_geometry_get(inst->win, &x, &y, &w, &h);
   //e_stolen_win_add(win);
   ecore_x_window_client_manage(win);
   /* Resize */
   ecore_x_window_resize(win, 48, 48);
   ecore_x_window_save_set_add(win);
   //ecore_x_window_shape_events_select(win, 1);
   //FIXME needed to resize the icon but freeze it ...
   //ecore_x_window_container_manage(ico->parent);
   ecore_x_window_reparent(ico->parent, inst->win, 
                           inst->padding.x + 48 * (num_ico - 1), 
                           inst->padding.y);
   ecore_x_window_reparent(win, ico->parent, 0, 0);
   //ecore_x_window_raise(inst->con->event_win);

   e_gadcon_client_min_size_set(inst->gcc, 48 * num_ico, 48);

   /* XXX trans hack 
    * Ok this is really ugly but (somehow) do the trick.
    * A copy of the icon size is kept inside the instance and duplicate as the
    * background pixmap of each icons. Which means that any non-linear
    * background will be wrongly faked behind tray icons.
    * We don't copy each time because the resize of the shelf may be done at the
    * same time than the XCopy and result in wrong pixmap.
    * An extra XWindow is added between the shelf and the tray icon.
    * That's because some app force the window pixmap to be null. And use the
    * parent bg pixmap/color to fake themselves the transparency.
    * TODO - Handle the case where the shelf is transparent
    */
   Ecore_X_Display *display;
   display = ecore_x_display_get();
   int wx, wy, ww, wh;
   GC context; 

   if (!inst->bg_pixmap)
     {
       context = ecore_x_gc_new(win);
       inst->bg_pixmap = ecore_x_pixmap_new(win, 48, 48, ecore_x_window_depth_get(win));
       XCopyArea(display, inst->win, inst->bg_pixmap, context, 
                 inst->padding.x, inst->padding.y, 48, 48, 0, 0);
       ecore_x_gc_del(context);
     }
   ecore_x_window_pixmap_set(ico->parent, inst->bg_pixmap);
   ecore_x_window_pixmap_set(ico->win, inst->bg_pixmap);

   ecore_x_window_show(ico->parent);
   ecore_x_window_show(win);
   XFlush(display);
}

static void 
_tray_set_active(void *data, int active) 
{
   Ecore_X_Window win;
   Display *disp;
   Ecore_X_Atom asel;
   char buf[32];
   Instance *inst = NULL;

   if (!(inst = data)) return;
   if (active) win = inst->win;
   disp = ecore_x_display_get();
   snprintf(buf, sizeof(buf), "_NET_SYSTEM_TRAY_S%d", inst->con->manager->num);
   asel = ecore_x_atom_get(buf);
   XSetSelectionOwner(disp, asel, win, CurrentTime);
   if (!active) return;
   if (XGetSelectionOwner(disp, asel) == inst->win) 
     {
	printf("Systray: Send owner Message\n");
	ecore_x_client_message32_send(inst->con->manager->root, 
				      ecore_x_atom_get("MANAGER"), 
				      ECORE_X_EVENT_MASK_WINDOW_CONFIGURE, 
				      CurrentTime, asel, win, 0, 0);
     }
}

static int 
_cb_msg(void *data, int type, void *event) 
{
   Ecore_X_Event_Client_Message *ev;
   Instance *inst = NULL;
   static Ecore_X_Atom aop = 0;
   static Ecore_X_Atom amsg = 0;
   static Ecore_X_Atom axmbed = 0;
   
   if (!(inst = data)) return 1;
   ev = event;
   printf("Systray: Got a message type : %i\n", ev->message_type);
   if (aop == 0)
     aop = ecore_x_atom_get("_NET_SYSTEM_TRAY_OPCODE");
   if (amsg = 0)
     amsg = ecore_x_atom_get("_NET_SYSTEM_TRAY_MESSAGE_DATA");
   if (axmbed == 0)
     axmbed = ecore_x_atom_get("_XEMBED");
   if (ev->message_type == aop) 
     {
	printf("Systray: Got OP Code\n");
	if (ev->data.l[1] == SYS_TRAY_REQUEST_DOCK) 
	  {
	     printf("Systray: Request Dock\n");
	     _tray_add_icon(inst, (Ecore_X_Window)ev->data.l[2]);
	     ecore_x_client_message32_send(ev->data.l[2], axmbed, 
					   ECORE_X_EVENT_MASK_NONE, CurrentTime, 
					   XEMBED_EMBEDDED_NOTIFY, 0, inst->win, 1);
          }
     }
   else if (ev->message_type == amsg)
     printf("Systray: Got MSG Code\n");
   else if (ev->message_type == axmbed)
     printf("Systray: Got XMBED Code\n");
   return 1;
}

static int 
_cb_destroy(void *data, int type, void *event) 
{
   Ecore_X_Event_Window_Destroy *ev;
   Instance *inst = NULL;
   int i;
   Evas_List *l;
   
   if (!(inst = data)) return 1;
   ev = event;
   for (l = inst->icons, i = 0; l; l = l->next, i++)
     {
       Tray_Icon *ico = l->data;
       if (ico->win == ev->win)
         {
           inst->icons = evas_list_remove_list(inst->icons, l);
           ecore_x_window_del(ico->parent);
           free(ico);
         }
       else
         {
           ecore_x_window_move(ico->parent,
                               inst->padding.x + 48 * i, 
                               inst->padding.y);
         }
     }
   return 1;
}
